/************************************************************************
 *
 *                    CalC (Calcium Calculator)
 *                         Victor Matveev
 *                        (GPL) 1999 - 2007 
 *                    LBM/NIDDK/NIH and DMS/NJIT
 *
 *                            calc.cpp
 *   The main() routine and the highest-order ADI engine procedures;
 *   tortuosity function interpreter/initializer
 *
 ************************************************************************/

#include "stdafx.h"
#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <string.h>
#include <stdarg.h>
#include "Ygl.h"
#include "syntax.h"
#include "box.h"
#include "grid.h"
#include "vector.h"
#include "field.h"
#include "table.h"
#include "peak.h"
#include "markov.h"
#include "interpol.h"
#include "simulation.h"
#include "fplot.h"
#include "gate.h"
#include "loop.h"

double CHARGE_LOSS;

int     VERBOSE = 2;
int     GEOMETRY = CARTESIAN3D;
int     GEOMETRY1 = 0;
int     GEOMETRY2 = 0;
int     GEOMETRY3 = 0;
char    LABEL_DIM1[2];
char    LABEL_DIM2[6];
char    LABEL_DIM3[4];
int     DIMENSIONALITY = 3;

FILE *stdlog = stderr;

//********************************************************************************************

#define  EXTRA_PARAM_STRING  "; pA=5.182134 ; pi=4 atan(1) ; "

#define RETURN_NORMAL 0     // return types used by ADI step methods
#define RETURN_END    1

//********************************************************************************************

void getRun( TokenString &TS, int i, bool *adaptive=0, char *PDEmethod=0, double *time=0, double *dt=0, 
             double *accuracy=0, double *dtStretch=0, double *ODEaccuracy=0);

const char *getMethod(char *smethod, CaMethod &, BufMethod &);
double get_sim_time(TokenString &params);

void setFieldByFunction(TokenString &TS, long p, double *Diff, const char *errStr);

int Buf1Dstep(BufferObj &Buf, VectorObj &Buf_next, FieldObj &Ca, double dt, double);
int Ca1Dstep(FieldObj &Ca, VectorObj &Canext, BufferArray &, double dt, double);

int Buf2Dstep(BufferObj &Buf, VectorObj &Buf_next, FieldObj &Ca, double dt, double);
int Ca2Dstep(FieldObj &Ca, VectorObj &Canext, BufferArray &, double dt, double);

int Buf2DFIstep(BufferObj &Buf, VectorObj &Buf_next, FieldObj &Ca, double dt, double);
int Ca2DFIstep(FieldObj &Ca, VectorObj &Canext, BufferArray &, double dt, double);

int Buf2DEulerStep(BufferObj &Buf, VectorObj &Buf_next, FieldObj &Ca, double dt, double);
int Ca2DEulerStep(FieldObj &Ca, VectorObj &Canext, BufferArray &, double dt, double);

int Ca3Dstep(FieldObj &Ca, VectorObj &Canext, BufferArray &, double dt, double);
int Buf3Dstep(BufferObj &Buf, VectorObj &Buf_next, FieldObj &Ca, double dt, double);

int Buf3DEulerStep(BufferObj &Buf, VectorObj &Buf_next, FieldObj &Ca, double dt, double);
int Ca3DEulerStep(FieldObj &Ca, VectorObj &Canext, BufferArray &, double dt, double);

int Ca3DFIstep(FieldObj &Ca, VectorObj &Canext, BufferArray &, double dt, double);
int Buf3DFIstep(BufferObj &Buf, VectorObj &Buf_next, FieldObj &Ca, double dt, double);

//**************************************************************************************************

void header() {
     fprintf(stdlog,"\n*************************************************************");
     fprintf(stdlog,"\n*                                                           *");
     fprintf(stdlog,"\n*  Calcium diffusion/buffering solver, version 5.7.9 (beta) *");
     fprintf(stdlog,"\n*                                                           *");
     fprintf(stdlog,"\n*            Victor Matveev (GPL) March 31, 2008            *");
     fprintf(stdlog,"\n* Dept of Math Sciences, New Jersey Institute of Technology *");
     fprintf(stdlog,"\n*                  and LBM, NIDDK, NIH                      *");
     fprintf(stdlog,"\n*                                                           *");
     fprintf(stdlog,"\n*      Supported in part by the NSF Grant DMS-0417416       *");
     fprintf(stdlog,"\n*                                                           *");
     fprintf(stdlog,"\n*************************************************************\n");
}

 //**************************************************************************************************
 //                                           M A I N
 //**************************************************************************************************


 int main(int argc, char **argv) {

 char script[512];

 try {

   if (argc < 2) {
     header();
     fprintf(stderr, "\n\nEnter the CalC script file name: ");
     scanf("%s", script);
   }
   else strcpy(script, argv[1]);

   TokenString *TS;

   TS = new TokenString(script, EXTRA_PARAM_STRING, "", argc, argv);
   TS->get_int_param("verbose", &VERBOSE);
   if (VERBOSE) header();

   VectorObj result( getTrackVarNum(*TS) );
   LoopObj   vary( *TS, &result );
   long seed = MY_SEED;
   seed =  TS->get_long_param("seed");
   srand48(seed);
   delete TS;

   // ++++++++++++++++++++++++++++ MAIN LOOP ++++++++++++++++++++++++++++

   for (int step = 0; step < vary.steps; step++) {
     vary.step();
     TS = new TokenString(script, EXTRA_PARAM_STRING, vary.loopVarString, argc, argv); 
     if (TS->token_count("exit")) { if (VERBOSE) fprintf(stderr, "\n > Exit command: breaking execution <\n "); break; }
     if (TS->token_count("continue")) { if (VERBOSE) fprintf(stderr, "\n > Continue command: breaking execution <\n "); delete TS; continue; }
     TS->get_int_param("verbose", &VERBOSE);
     SimulationObj *Simulation;
     double *trackPtr;

     if ( !TS->token_count("Run") ) {                    //####### CALCULATOR MODE #######
       if (VERBOSE) fprintf(stdlog, "***** No simulation runs specified; running in calculator mode *****\n");
       if ( vary() ) for (int i = 0; i < result.size; i++) 
	  result[i] = TS->get_double( getTrackVar(*TS, i) );
       TS->printResults(0);
     }        
     else {
       if ( !(TS->token_count("Ca.D") || TS->token_count("grid")) || equal(TS->get_string("mode"), "ODE") ) {
         if (VERBOSE) fprintf(stdlog, "***** Running in ODE-only mode *****\n");
	 Simulation = new ODESimulationObj(*TS);    //#######  ODE solver mode: ######
       } 
       else  Simulation = new SimulationObj(*TS);       //#######  FULL PDE MODE  #######
       Simulation->Run();                               //##### RUN THE DIFFERENCE SCHEME #####
       if ( vary() ) for (int i = 0; i < result.size; i++) 
         if ( (trackPtr = Simulation->ResolveID( vary.trackIDs[i] )) ) result[i] = *trackPtr;
         else  TS->errorMessage( TS->token_index(vary.trackIDs[i]), 0, "Cannot track an undefined variable" );
       TS->printResults(Simulation);
       delete Simulation;
       }
     vary.draw(); 
     delete TS;
   }

 }
 catch (char *str) {  if ( !equal(str,"") ) {
                        fprintf(stdlog, "\n\n*** Error: %s\n", str);
                        delete [] str;  }
                   }
 catch (int i) { i = 0; }

 return 0;
 }


 //**********************************************************************************************
 //@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
 //**********************************************************************************************

//**************************************************************************
//                C A L C I U M   D - G   S T E P
//**************************************************************************
//
//  (1 - Ax/2) Ca*  = (1 + Ax/2 + Ay + Az) Ca + dt H(Ca,B)
//  (1 - Ay/2) Ca** = -Ay/2 Ca + Ca*
//  (1 - Az/2) Ca^  = -Az/2 Ca + Ca** + dt / 2 * { H(Ca^,B) - H(Ca,B) }
//
//  H(Ca,B) = Sum_B{ - kplus * Ca * B + kminus * (B_total - B) }
//
//  Ca = Ca(t_n), Ca^ = Ca(t_{n+1}), B = B(t_{n+1/2})
//**************************************************************************

int Ca3Dstep(FieldObj &Ca, VectorObj &Next, BufferArray &Buf, double dt, double T)
{
  double    t0 = Ca.Time;
  int       rvalue = RETURN_NORMAL;
  VectorObj Lin(FieldObj::Size);
 
  if ( (T > 0.0) && (t0 + dt > T))  {  rvalue = RETURN_END; dt = T - t0; }

  double nu = Ca.getD() * dt;
  Next = Ca;
  Lin = 0.0;
  for (int b = 0; b < Buf.buf_num; b++)
    {
    Lin  -= (dt * Buf.array[b]->kplus) * *Buf.array[b];
    Next += ( dt * Buf.array[b]->kminus) * (*(Buf.array[b]->total) - *Buf.array[b]);
    }

  Next += Lin % Ca;
  Ca.Time += 0.5 * dt;
  Ca.add_sources(Next, dt);
  /*
  Ca.Run3Dx(-0.5 * nu, 0.5 * nu,    nu,     nu,  0,  Next.elem);
  Ca.Run3Dy(-0.5 * nu,     0.,  -0.5 * nu,  0.,  0,  Next.elem);

  Lin *= 0.5;
  Next -= (Lin % Ca);

  Ca.Run3Dz(-0.5 * nu, 0., 0., -0.5 * nu, Lin.elem, Next.elem);
  */
  Ca.Run3Dy(-0.5 * nu,     nu,  0.5*nu,  nu,      0,  Next.elem);
  Ca.Run3Dz(-0.5 * nu,     0.,  0.0,   -0.5 * nu, 0,  Next.elem);

  Lin *= 0.5;
  Next -= (Lin % Ca);

  Ca.Run3Dx(-0.5 * nu, -0.5 * nu, 0.0, 0.0, Lin.elem, Next.elem);
  if (T > 0.0) { Ca = Next; Ca.Time += 0.5 * dt; } /*Ca.cleanBoundaries();*/ 
  else Ca.Time = t0;

  return rvalue;
}

//**************************************************************************
//                  B U F F E R   D - G   S T E P
//**************************************************************************

int Buf3Dstep(BufferObj &Buf, VectorObj &Next, FieldObj &Ca, double dt, double T)
{ 
int    rvalue = RETURN_NORMAL;

  if ( (T > 0.0) && (Buf.Time + dt > T)) 
    { dt = T - Buf.Time; /*printf("Buf: dt reduced to %g\n", dt);*/ rvalue = RETURN_END; }
  double nu = Buf.getD() * dt;
  
  VectorObj Lin( -dt * (Buf.kplus * Ca + Buf.kminus ) );
  Next = Buf % ( 1.0 + Lin ) + dt * Buf.kminus * *(Buf.total);
  
  Buf.Run3Dx(-0.5 * nu, 0.5 * nu, nu, nu, 0, Next.elem);
  Buf.Run3Dy(-0.5 * nu, 0., -0.5 * nu, 0., 0, Next.elem);

  Lin *= 0.5;
  Next -= (Lin % Buf);
  Buf.Run3Dz(-0.5 * nu, 0., 0., -0.5 * nu,  Lin.elem, Next.elem);

 if (T > 0.0) {  Buf.Time += dt; Buf = Next; } /*Buf.cleanBoundaries();*/ 
  
  return rvalue;
}

//**************************************************************************
//                C A L C I U M   2 D   D - G   S T E P
//**************************************************************************
//
//  (1 - Ax/2) Ca*  = (1 + Ay/2) Ca  + dt/2 H(Ca,B)
//  (1 - Ay/2) Ca^  = (1 + Ax/2) Ca* + dt/2 H(Ca^,B)
//
//  H(Ca,B) = Sum_B{ - kplus * Ca * B + kminus * (B_total - B) }
//
//  Ca = Ca(t_n), Ca^ = Ca(t_{n+1}), B = B(t_{n+1/2})
//**************************************************************************

int Ca2Dstep(FieldObj &Ca, VectorObj &Next, BufferArray &Buf, double dt, double T)
{
  double    t0 = Ca.Time;
  int       b, rvalue = RETURN_NORMAL;
  VectorObj Lin(FieldObj::Size);
  FieldObj  Temp(Ca);

  if ( (T > 0.0) && (t0 + dt > T))  { rvalue = RETURN_END; dt = T - t0; }

  double nu = Ca.getD() * dt;

  Next = Ca;
  Lin = 0.0; 

  for (b = 0; b < Buf.buf_num; b++)
    {
    Lin  -= (0.5 * dt * Buf.array[b]->kplus) * *Buf.array[b];
    Next += (*(Buf.array[b]->total) - *Buf.array[b]) * (0.5 * dt * Buf.array[b]->kminus);
    }

  Next += Lin % Ca;
  Ca.Time += 0.5 * dt;
  Ca.add_sources(Next, 0.5 * dt);

  Ca.Run2Dx(    -0.5 * nu,     0.0,    0.5 * nu,  0,  Next.elem);

  Temp = Next; // IMPORTANT! cf. 3D methods.
  for (b = 0; b < Buf.buf_num; b++)    
    Next += (*(Buf.array[b]->total) - *Buf.array[b]) * (0.5 * dt * Buf.array[b]->kminus);
  Ca.add_sources(Next, 0.5 * dt);

  Temp.Run2Dy(-0.5 * nu, 0.5 * nu,     0.0,     Lin.elem,  Next.elem);

  if (T > 0.0) { Ca = Next; Ca.Time += 0.5 * dt; } /*Ca.cleanBoundaries();*/ 
  else Ca.Time = t0;
  return rvalue;
}

//**************************************************************************
//               B U F F E R   2 D   D - G   S T E P
//**************************************************************************
//
//  (1 - Ax/2) B*  = (1 + Ay/2) B  + dt/2 H(Ca,B)
//  (1 - Ay/2) B^  = (1 + Ax/2) B* + dt/2 H(Ca,B^)
//
//  H(Ca,B) = - kplus * Ca * B + kminus * (B_total - B)
//
//  B = B(t_n), B^ = B(t_{n+1}), Ca = Ca(t_{n+1/2})
//**************************************************************************

int Buf2Dstep(BufferObj &Buf, VectorObj &Next, FieldObj &Ca, double dt, double T)
{ 
int       rvalue = RETURN_NORMAL;
FieldObj  Temp(Buf);

  if ( (T > 0.0) && (Buf.Time + dt > T)) 
    { dt = T - Buf.Time; /*printf("Buf: dt reduced to %g\n", dt);*/ rvalue = RETURN_END; }
  double nu = Buf.getD() * dt;
  
  VectorObj Lin( (-0.5 * dt) * (Buf.kplus * Ca + Buf.kminus) );

  Next = Buf + 0.5 * dt * Buf.kminus * *(Buf.total) + Lin % Buf;

  Buf.Run2Dx(    -0.5 * nu,     0.0,    0.5 * nu,  0,  Next.elem);

  Temp = Next; // IMPORTANT! cf. 3D methods.
  Next += 0.5 * dt * Buf.kminus * *Buf.total;
  
  Temp.Run2Dy(-0.5 * nu, 0.5 * nu,     0.0,     Lin.elem,  Next.elem);

  if (T > 0.0) { Buf.Time += dt; Buf = Next; } /*Buf.cleanBoundaries();*/ 
  return rvalue;
}

//**************************************************************************
//                C A L C I U M   2 D   F I   S T E P
//**************************************************************************
//
//  (1 - Ax) Ca*  = (1 + Ay) Ca + dt H(Ca,B)
//  (1 - Ay) Ca^  = Ca* - Ay Ca + dt ( H(Ca^,B) - H(Ca,B) )
//
//  H(Ca,B) = Sum_B{ - kplus * Ca * B + kminus * (B_total - B) } + dt I
//
//  Ca = Ca(t_n), Ca^ = Ca(t_{n+1}), B = B(t_{n+1/2})
//**************************************************************************

int Ca2DFIstep(FieldObj &Ca, VectorObj &Next, BufferArray &Buf, double dt, double T)
{
  double    t0 = Ca.Time;
  int       b, rvalue = RETURN_NORMAL;
  VectorObj Lin(FieldObj::Size);
 
  if ( (T > 0.0) && (t0 + dt > T))  { rvalue = RETURN_END; dt = T - t0; }

  double nu = Ca.getD() * dt;

  Next = Ca;
  Lin = 0.0; 

  for (b = 0; b < Buf.buf_num; b++)
    {
    Lin  -= (dt * Buf.array[b]->kplus) * *Buf.array[b];
    Next += (*(Buf.array[b]->total) - *Buf.array[b]) * (dt * Buf.array[b]->kminus);
    }

  Next += Lin % Ca;
  Ca.Time += 0.5 * dt;
  Ca.add_sources(Next, dt);

  Ca.Run2Dx( -nu,  0.0,   nu,  0,  Next.elem);
  Next -= Lin % Ca;

  Ca.Run2Dy(-nu, 0.0, -nu,  Lin.elem,  Next.elem);

  if (T > 0.0) { Ca = Next; Ca.Time += 0.5 * dt; }/*Ca.cleanBoundaries();*/ 
  else Ca.Time = t0;
  return rvalue;
}

//**************************************************************************
//               B U F F E R   2 D   F I   S T E P
//**************************************************************************
//
//  (1 - Ax) B*  = (1 + Ay) B + dt H(Ca,B)
//  (1 - Ay) B^  = B* - Ay B  + dt ( H(Ca,B^) - H(Ca,B) )
//
//  H(Ca,B) = - kplus * Ca * B + kminus * (B_total - B)
//
//  B = B(t_n), B^ = B(t_{n+1}), Ca = Ca(t_{n+1/2})
//**************************************************************************

int Buf2DFIstep(BufferObj &Buf, VectorObj &Next, FieldObj &Ca, double dt, double T)
{ 
int    rvalue = RETURN_NORMAL;

  if ( (T > 0.0) && (Buf.Time + dt > T))  { dt = T - Buf.Time; rvalue = RETURN_END; }
  double nu = Buf.getD() * dt;
  
  VectorObj Lin( (-dt) * (Buf.kplus * Ca + Buf.kminus) );

  Next = Buf + dt * Buf.kminus * *Buf.total + Lin % Buf;

  Buf.Run2Dx(    -nu,     0.0,   nu,  0,  Next.elem);

  Next -= Lin % Buf;
  
  Buf.Run2Dy(-nu, 0.0, -nu,  Lin.elem,  Next.elem); 

  if (T > 0.0) { Buf.Time += dt; Buf = Next; } /*Buf.cleanBoundaries();*/  
  return rvalue;
}

//**************************************************************************
//         C A L C I U M   F U L L Y   I M P L I C I T    S T E P
//**************************************************************************
//
//  (1 - Ax) Ca*  = (1 + Ay + Az) Ca + dt H(Ca,B)
//  (1 - Ay) Ca** = -Ay Ca + Ca*
//  (1 - Az) Ca^  = -Az Ca + Ca** + dt * { H(Ca^,B) - H(Ca,B) }  - no 1/2 !
//
//  H(Ca,B) = Sum_B{ - kplus * Ca * B + kminus * (B_total - B) }
//
//  Ca = Ca(t_n), Ca^ = Ca(t_{n+1}), B = B(t_{n+1/2})
//**************************************************************************

int Ca3DFIstep(FieldObj &Ca, VectorObj &Next, BufferArray &Buf, double dt, double T)
{
  double    t0 = Ca.Time;
  int       rvalue = RETURN_NORMAL;
  VectorObj Lin(FieldObj::Size);
 
  if ( (T > 0.0) && (t0 + dt > T))  {  rvalue = RETURN_END; dt = T - t0; }

  double nu = Ca.getD() * dt;
  Next = Ca;
  Lin = 0.0;
  for (int b = 0; b < Buf.buf_num; b++)
    {
    Lin  -= (dt * Buf.array[b]->kplus) * *Buf.array[b];
    Next += ( dt * Buf.array[b]->kminus) * (*(Buf.array[b]->total) - *Buf.array[b]);
    }

  Next += Lin % Ca;
  Ca.Time += 0.5 * dt;
  Ca.add_sources(Next, dt);

  Ca.Run3Dx(-nu,  0,   nu,  nu,  0,  Next.elem);
  Ca.Run3Dy(-nu,  0., -nu,  0.,  0,  Next.elem);
  Next -= (Lin % Ca);

  Ca.Run3Dz(-nu, 0., 0., -nu, Lin.elem, Next.elem);
  if (T > 0.0) { Ca = Next; Ca.Time += 0.5 * dt; } /*Ca.cleanBoundaries();*/ 
  else Ca.Time = t0;

  return rvalue;
}

//******************************************************************************************
//                B U F F E R   F U L L Y   I M P L I C I T    S T E P
//******************************************************************************************
//
//  (1 - Ax) Buf*  = (1 + Ay + Az) Buf + dt H(Buf,B)
//  (1 - Ay) Buf** = -Ay Buf + Buf*
//  (1 - Az) Buf^  = -Az Buf + Buf** + dt * { H(Buf^,Ca) - H(Buf,Ca) }  - no 1/2 !
//
//  H(Buf,B) = - kplus * Buf * Ca + kminus * (B_total - Buf) }
//
//  Buf = Buf(t_n), Buf^ = Buf(t_{n+1}), Ca = Ca(t_{n+1/2})
//**************************************************************************

int Buf3DFIstep(BufferObj &Buf, VectorObj &Next, FieldObj &Ca, double dt, double T)
{ 
int    rvalue = RETURN_NORMAL;

  if ( (T > 0.0) && (Buf.Time + dt > T)) 
    { dt = T - Buf.Time; /*printf("Buf: dt reduced to %g\n", dt);*/ rvalue = RETURN_END; }
  double nu = Buf.getD() * dt;
  
  VectorObj Lin( -dt * (Buf.kplus * Ca + Buf.kminus ) );
  Next = Buf % ( 1.0 + Lin ) + dt * Buf.kminus * *(Buf.total);

  Buf.Run3Dx(-nu, 0.0, nu, nu, 0, Next.elem);
  Buf.Run3Dy(-nu, 0., -nu, 0., 0, Next.elem);

  Next -= (Lin % Buf);
  Buf.Run3Dz(-nu, 0., 0., -nu,  Lin.elem, Next.elem);

 if (T > 0.0) {  Buf.Time += dt; Buf = Next; } /*Buf.cleanBoundaries();*/ 
  
  return rvalue;
}


//**************************************************************************
//              C A L C I U M   E U L E R   S T E P
//**************************************************************************
//
//  Ca(t_{n+1}) = (1 + Ax + Ay + Az) Ca(t_n) + dt H(Ca(t_n), B(t_n))
//
//  H(Ca,B) = Sum_B{ - kplus * Ca * B + kminus * (B_total - B) }
//
//**************************************************************************

int Ca3DEulerStep(FieldObj &Ca, VectorObj &Next, BufferArray &Buf, double dt, double T)
{
  int    rvalue = RETURN_NORMAL;
  double    t0 = Ca.Time;

  if ( (T > 0.0) && (t0 + dt > T)) { dt = T - t0; rvalue = RETURN_END; }
  double nu = Ca.getD() * dt;

  Next = Ca;     // take care of non-updating values at Dirichlet boundary
  for (int b = 0; b < Buf.buf_num; b++)
    Next -= *Buf.array[b] % ( dt * Buf.array[b]->kminus + (dt * Buf.array[b]->kplus) * Ca ) +
            *(Buf.array[b]->total) * dt * Buf.array[b]->kminus;

  Ca.Time += 0.5 * dt;
  Ca.add_sources(Next, dt);
  Ca.Run3Dx(0.0, nu, nu, nu, 0, Next.elem);

  if (T > 0.0) { Ca = Next; Ca.Time += 0.5 * dt; /*Ca.cleanBoundaries();*/ }
  else Ca.Time = t0;

  return rvalue;
}

//**************************************************************************
//            B U F F E R   E U L E R   S T E P
//**************************************************************************

int Buf3DEulerStep(BufferObj &Buf, VectorObj &Next, FieldObj &Ca, double dt, double T)
{
  int    rvalue = RETURN_NORMAL;

  if ( (T > 0.0) && (Buf.Time + dt > T)) { dt = T - Buf.Time; rvalue = RETURN_END; }
  double nu = Buf.getD() * dt;

  Next = Buf % ( 1.0 - dt * Buf.kminus - (dt * Buf.kplus) * Ca ) + dt * Buf.kminus * *Buf.total;
 
  Buf.Run3Dx(0.0, nu, nu, nu, 0, Next.elem);
  if (T > 0.0) { Buf = Next; Buf.Time += dt; } /*Buf.cleanBoundaries();*/

  return rvalue;
}

//**************************************************************************
//              C A L C I U M   2 D   E U L E R   S T E P
//**************************************************************************
//
//  Ca(t_{n+1}) = (1 + Ax + Ay) Ca(t_n) + dt H(Ca(t_n), B(t_n))
//
//  H(Ca,B) = Sum_B{ - kplus * Ca * B + kminus * (B_total - B) }
//
//**************************************************************************

int Ca2DEulerStep(FieldObj &Ca, VectorObj &Next, BufferArray &Buf, double dt, double T)
{
  int    rvalue = RETURN_NORMAL;
  double t0 = Ca.Time;

  if ( (T > 0.0) && (t0 + dt > T)) { dt = T - t0; rvalue = RETURN_END; }
  double nu = Ca.getD() * dt;

  Next = Ca;     // take care of non-updating values at Dirichlet boundary
  for (int b = 0; b < Buf.buf_num; b++)
    Next -= *Buf.array[b] % ( dt * Buf.array[b]->kminus + (dt * Buf.array[b]->kplus) * Ca )
         +  *(Buf.array[b]->total) * dt * Buf.array[b]->kminus;

  Ca.Time += 0.5 * dt;
  Ca.add_sources(Next, dt);

  Ca.Run2Dy(0.0, nu, nu, 0, Next.elem);

  if (T > 0.0) { Ca = Next; Ca.Time += 0.5 * dt; /*Ca.cleanBoundaries();*/ } else Ca.Time = t0;

  return rvalue;
}

//**************************************************************************
//            B U F F E R   2 D   E U L E R   S T E P
//**************************************************************************

int Buf2DEulerStep(BufferObj &Buf, VectorObj &Next, FieldObj &Ca, double dt, double T)
{
  int    rvalue = RETURN_NORMAL;

  if ( (T > 0.0) && (Buf.Time + dt > T)) { dt = T - Buf.Time; rvalue = RETURN_END; }
  double nu = Buf.getD() * dt;

  Next = Buf % ( 1.0 - dt * Buf.kminus - (dt * Buf.kplus) * Ca ) + dt * Buf.kminus * *Buf.total;
 
  Buf.Run2Dx(0.0, nu, nu, 0, Next.elem);

  if (T > 0.0) { Buf = Next; Buf.Time += dt; /*Buf.cleanBoundaries();*/ }

  return rvalue;
}

/**************************************************************************
     C A L C I U M   C R A N K - N I C H O L S O N   1 D   S T E P
***************************************************************************

  (1 - Ax/2) Ca^ = (1 + Ax/2) Ca + dt/2 ( H(Ca, B) + H(Ca^,B) ) + dt sources(t_{n+1/2})
                 = (1 + Ax/2) Ca - dt/2 Sum_B{ kplus * B } ( Ca + Ca^ )
                 + dt { Sum_B{ kminus * (B_total - B) + sources(t_{n+1/2}) }

  ( 1 - Ax/2 - Lin ) Ca^ = Ax/2 Ca + (1 + Lin ) Ca
                         + dt { Sum_B{ kminus * (B_total - B) + sources(t_{n+1/2}) }

  Lin = - dt/2 Sum_B{ kplus * B }


  H(Ca,B) = Sum_B{ - kplus * Ca * B + kminus * (B_total - B) }

  Ca = Ca(t_n), Ca^ = Ca(t_{n+1}), B = B(t_{n+1/2})
****************************************************************************/

int Ca1Dstep(FieldObj &Ca, VectorObj &Next, BufferArray &Buf, double dt, double T)
{
  int       rvalue = RETURN_NORMAL;
  double    t0 = Ca.Time;
  VectorObj Lin(FieldObj::Size);


  if ( (T > 0.0) && (t0 + dt > T))  { rvalue = RETURN_END; dt = T - t0; }

  double nu = Ca.getD() * dt;

  Next = Ca;

  Lin = 0.0; 
  for (int b = 0; b < Buf.buf_num; b++) {
    Lin  -= (0.5 * dt * Buf.array[b]->kplus) * *Buf.array[b];
    Next += ( dt * Buf.array[b]->kminus ) * (*(Buf.array[b]->total) - *Buf.array[b]);
    }

  Next += Ca % Lin;
  Ca.Time += 0.5 * dt;
  Ca.add_sources(Next, dt);
  Ca.Run1D( -0.5 * nu, 0.5 * nu, Lin.elem, Next.elem);

  if (T > 0.0) { Ca = Next; Ca.Time += 0.5 * dt; /*Ca.cleanBoundaries();*/ }
    else Ca.Time = t0;

  return rvalue;
}

/**************************************************************************
       B U F F E R    C R A N K - N I C H O L S O N   1 D   S T E P
**************************************************************************

  (1 - Ax/2) B^ = (1 + Ax/2) B + dt/2 ( H(Ca, B) + H(Ca,B^) )
                = (1 + Ax/2) B - dt/2 (kplus Ca + kminus) ( B + B^ ) + dt kminus B_total 

  ( 1 - Ax/2 - Lin ) B^ = Ax/2 Ca + (1 + Lin ) B + dt kminus B_total

  Lin = - dt/2 (kplus Ca + kminus)

  H(Ca,B) = - kplus * Ca * B + kminus * (B_total - B)

  B = B(t_n), B^ = B(t_{n+1}), Ca = Ca(t_{n+1/2})
****************************************************************************/

int Buf1Dstep(BufferObj &Buf, VectorObj &Next, FieldObj &Ca, double dt, double T)
{
  int    rvalue = RETURN_NORMAL;

  if ( (T > 0.0) && (Buf.Time + dt > T)) { dt = T - Buf.Time; rvalue = RETURN_END; }
  double nu = Buf.getD() * dt;

  VectorObj Lin( (-0.5 * dt) * (Buf.kplus * Ca + Buf.kminus ) );
  Next = Buf % ( 1.0 + Lin ) + (dt * Buf.kminus) * *Buf.total;

  Buf.Run1D( -0.5 * nu, 0.5 * nu, Lin.elem, Next.elem);

  if (T > 0.0) { Buf = Next; Buf.Time += dt; /*Buf.cleanBoundaries();*/ }

  return rvalue;
}

//**************************************************************************

const char *getMethod( char *smethod, CaMethod &CaStep, BufMethod &BufStep) {

 switch (DIMENSIONALITY)  { 

 case 1:  
     if ( equal(smethod,"CN") || equal(smethod,"default") ) 
       { CaStep = &Ca1Dstep; BufStep = &Buf1Dstep; return makeMessage("1D (%s) Crank-Nicholson implicit", LABEL_DIM1); }
     break;

 case 2:
     if ( equal(smethod,"DG") || equal(smethod,"default") )
       { CaStep = &Ca2Dstep; BufStep = &Buf2Dstep; return  makeMessage("2D (%s,%s) Crank-Nicholson implicit", LABEL_DIM1, LABEL_DIM2); }
     else  if ( equal(smethod,"EU") )
       { CaStep = &Ca2DEulerStep; BufStep = &Buf2DEulerStep; return makeMessage("2D (%s,%s) Euler", LABEL_DIM1, LABEL_DIM2); }
     else  if ( equal(smethod,"FI") )
       { CaStep = &Ca2DFIstep; BufStep = &Buf2DFIstep; return makeMessage("2D (%s,%s) fully implicit", LABEL_DIM1, LABEL_DIM2); }
     break;
 case 3:  
     if ( equal(smethod,"FI") )
       { CaStep = &Ca3DFIstep; BufStep = &Buf3DFIstep;   return  makeMessage("3D (%s,%s,%s) fully implicit", LABEL_DIM1, LABEL_DIM2, LABEL_DIM3); }
     else if ( equal(smethod,"EU") )
       { CaStep = &Ca3DEulerStep; BufStep = &Buf3DEulerStep;   return  makeMessage("3D (%s,%s,%s) Euler", LABEL_DIM1, LABEL_DIM2, LABEL_DIM3); }
     else if ( equal(smethod,"DG") || equal(smethod,"default") ) // default
       { CaStep = &Ca3Dstep; BufStep = &Buf3Dstep;       return  makeMessage("3D (%s,%s,%s) Douglas-Gunn ADI", LABEL_DIM1, LABEL_DIM2, LABEL_DIM3); }
     break;

 }

 throw makeMessage("Cannot recognize difference scheme \"%s\"", smethod);
}

//@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@


void getRun( TokenString &TS, int i, bool *adaptive, char *PDEmethod, double *time, double *dt,
     double *accuracy, double *dtStretch, double *ODEaccuracy)
{
  char firstArg[MAX_TOKEN_LENGTH], secondArg[MAX_TOKEN_LENGTH];

  if (adaptive) *adaptive = false;             // non-adaptive is the default
  long pos = TS.token_index("Run", i);

  TS.trail_pars(pos, 's', firstArg, 's', secondArg, 'E');
  pos ++;
  int pmax = TS.tokens_to_eol(pos);

  if ( pmax > 0 && !TS.isConst(firstArg) ) { 
    pos++; 
    if (adaptive) if ( equal(firstArg,"adaptive") ) *adaptive = true;
                  else if (PDEmethod) strcpy(PDEmethod, firstArg);
    else if (PDEmethod) strcpy(PDEmethod, firstArg);

    if ( pmax > 1 && !TS.isConst(secondArg) ) { 
         pos++; 
         if (PDEmethod) strcpy(PDEmethod, secondArg);
    }
  }

  if (time) {
    *time = TS.get_double(pos);
    if (*time <= 0) TS.errorMessage(pos, makeMessage("Simulation time has to be positive (T = %lf)", *time) );
  } 

  TS.trail_pars(pos, 'd', dt, 'd', accuracy, 'd', dtStretch, 'd', ODEaccuracy, 'E');

  if (dt) if (*dt <= 0) 
    TS.errorMessage(pos+1, makeMessage("Time step must be positive (dt = %lf)", *dt) );
  if (accuracy) if (*accuracy <= 0 || *accuracy > 1.0) // this argument is actually ODEaccuracy for non-adaptive run
    TS.errorMessage(pos+2, makeMessage("Accuracy must be between 0 and 1 (accuracy = %lf)", *accuracy) );
  if (dtStretch) if (*dtStretch < 1) 
    TS.errorMessage(pos+3, makeMessage("Time step stretch must be greater than 1 (dtStretch = %lf)", *dtStretch) );
  if (ODEaccuracy) if (*ODEaccuracy <= 0 || *ODEaccuracy > 1.0) 
    TS.errorMessage(pos+4, makeMessage("ODE accuracy must be between 0 and 1 (ODEaccuracy = %lf)", *ODEaccuracy) );
}


//**************************************************************************

double get_sim_time(TokenString &params)
  {
  double tt, t = 0.0;

  for (int i = 1; i<= params.token_count("Run"); i++) {
    getRun(params, i, 0, 0, &tt); 
    t += tt;
  }

  return t;
  }


//**************************************************************************


void setFieldByFunction(TokenString &TS, long p, double *Field, const char *errStr) {

  int xsize = FieldObj::Grid->xsize;  double *xcoord = FieldObj::Grid->xcoord;
  int ysize = FieldObj::Grid->ysize;  double *ycoord = FieldObj::Grid->ycoord;
  int zsize = FieldObj::Grid->zsize;  double *zcoord = FieldObj::Grid->zcoord;

  int  ix, iy, iz;
  long l = 0;
  double x, y, z;
  ExpressionObj *T = 0;

  switch (DIMENSIONALITY) {

  case 1:
    T = new ExpressionObj(TS, p, makeMessage("%s (function of %s)", errStr, LABEL_DIM1), 0, 0, &x, LABEL_DIM1); 
    for (ix = 0; ix < xsize; ix++)  {
          x = xcoord[ix];
          Field[l++] = T->Evaluate();
    }
    break;  

  case 2:
    T = new ExpressionObj(TS, p, 
            makeMessage("%s (function of %s, %s)", errStr, LABEL_DIM1, LABEL_DIM2), 0, 0, &x, LABEL_DIM1, &y, LABEL_DIM2); 
    for (iy = 0; iy < ysize; iy++) {
        y = ycoord[iy];
        for (ix = 0; ix < xsize; ix++)  {
          x = xcoord[ix];
          Field[l++] = T->Evaluate();
	}
    }
    break;  
 
  case 3:
    T = new ExpressionObj(TS, p, 
         makeMessage("%s (function of %s, %s, %s)", errStr, LABEL_DIM1, LABEL_DIM2, LABEL_DIM3), 
         0, 0, &x, LABEL_DIM1, &y, LABEL_DIM2, &z, LABEL_DIM3); 
    for (iz = 0; iz < zsize; iz++) {
      z = zcoord[iz];
      for (iy = 0; iy < ysize; iy++) {
        y = ycoord[iy];
        for (ix = 0; ix < xsize; ix++)  {
          x = xcoord[ix];
          Field[l++] = T->Evaluate();
	}
      }
    }
    break;  

 }

 if (T && VERBOSE) T->print(stdlog); 
 if (T) delete T;
 return;
}


//**************************************************************************
